/*
/ TextCsv.cpp
/ methods related to CSV/TXT loading 
/
/ version 1.2, 2008 October 9
/
/ Author: Sandro Furieri a-furieri@lqt.it
/
/ Copyright (C) 2008  Alessandro Furieri
/
/    This program is free software: you can redistribute it and/or modify
/    it under the terms of the GNU General Public License as published by
/    the Free Software Foundation, either version 3 of the License, or
/    (at your option) any later version.
/
/    This program is distributed in the hope that it will be useful,
/    but WITHOUT ANY WARRANTY; without even the implied warranty of
/    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/    GNU General Public License for more details.
/
/    You should have received a copy of the GNU General Public License
/    along with this program.  If not, see <http://www.gnu.org/licenses/>.
/
*/

#include "Classdef.h"

#if defined(_WIN32) && !defined(__MINGW32__)
#define strcasecmp	_stricmp
#endif

#define VRTTXT_TEXT		1
#define VRTTXT_INTEGER	2
#define VRTTXT_DOUBLE	3

#define VTXT_ROW_BLOCK	65536

struct row_buffer
{
/* a complete row */
  int n_cells;                  /* how many cells are stored into this line */
  char **cells;                 /* the cells array */
  char *text;                   /* the text-cells array */
  struct row_buffer *next;      /* pointer for linked list */
};

struct row_pointer
{
/* a row pointer */
  long offset;                  /* the row start offset */
  int len;                      /* the row length */
  char start;                   /* the first char of the row - signature */
  char valid;                   /* 1=valid - 0=invalid */
  int n_cells;                  /* # cells */
};

struct row_block
{
/* a block of row pointers */
  int n_rows;
  struct row_pointer rows[VTXT_ROW_BLOCK];
  struct row_block *next;
};

struct text_buffer
{
  FILE *text_file;              /* the underlaying file */
  void *toUtf8;                 /* the UTF-8 ICONV converter */
  char field_separator;
  char text_separator;
  char decimal_separator;
  int max_n_cells;              /* the maximun cell index */
  char **titles;                /* the column titles array */
  char *types;                  /* the column types array */
  int n_rows;                   /* the number of rows */
  struct row_pointer **rows;    /* the rows array */
  struct row_block *first;      /* pointers to build a linked list of rows */
  struct row_block *last;
  struct row_buffer *current_row_buffer;
};

static int text_add_block(struct text_buffer *text)
{
/* inserting a block of rows into the text buffer struct */
  struct row_block *block =
    (struct row_block *) malloc(sizeof(struct row_block));
  if (block == NULL)
    return 0;
  block->n_rows = 0;
  block->next = NULL;
/* inserting the block of rows into the linked list */
  if (!(text->first))
    text->first = block;
  if (text->last)
    text->last->next = block;
  text->last = block;
  return 1;
}

static struct row_pointer *get_row_pointer(struct text_buffer *text,
                                           long current_row)
{
/* retrieving a ROW POINTER struct */
  if (!text)
    return NULL;
  if (!(text->rows))
    return NULL;
  if (current_row < 0 || current_row >= text->n_rows)
    return NULL;
  return *(text->rows + current_row);
}

static int
text_insert_row(struct text_buffer *text, long offset, int len, char start)
{
/* inserting a row into the text buffer struct */
  struct row_pointer *row;
  if (text->last == NULL)
    {
      if (!text_add_block(text))
        return 0;
    }
  if (text->last->n_rows == VTXT_ROW_BLOCK)
    {
      if (!text_add_block(text))
        return 0;
    }
  row = text->last->rows + text->last->n_rows;
  row->offset = offset;
  row->len = len;
  row->start = start;
  row->valid = 0;
  row->n_cells = 0;
  (text->last->n_rows)++;
  return 1;
}

static struct text_buffer *text_buffer_alloc(FILE * fl, void *toUtf8,
                                             char field_separator,
                                             char text_separator,
                                             char decimal_separator)
{
/* allocating and initializing the text buffer struct */
  struct text_buffer *text =
    (struct text_buffer *) malloc(sizeof(struct text_buffer));
  if (text == NULL)
    return NULL;
  text->text_file = fl;
  text->toUtf8 = toUtf8;
  text->field_separator = field_separator;
  text->text_separator = text_separator;
  text->decimal_separator = decimal_separator;
  text->max_n_cells = 0;
  text->titles = NULL;
  text->types = NULL;
  text->n_rows = 0;
  text->rows = NULL;
  text->first = NULL;
  text->last = NULL;
  text->current_row_buffer = NULL;
  return text;
}

static void text_row_free(struct row_buffer *row)
{
/* memory cleanup - freeing a row buffer */
  int i;
  if (!row)
    return;
  if (row->cells)
    {
      for (i = 0; i < row->n_cells; i++)
        {
          if (*(row->cells + i))
            {
              free(*(row->cells + i));
            }
        }
      free(row->cells);
    }
  if (row->text)
    free(row->text);
  free(row);
}

static void text_buffer_free(struct text_buffer *text)
{
/* memory cleanup - freeing the text buffer */
  int i;
  struct row_block *block;
  struct row_block *pn;
  if (!text)
    return;
  block = text->first;
  while (block)
    {
      pn = block->next;
      free(block);
      block = pn;
    }
  if (text->titles)
    {
      for (i = 0; i < text->max_n_cells; i++)
        {
          if (*(text->titles + i))
            free(*(text->titles + i));
        }
      free(text->titles);
    }
  if (text->rows)
    free(text->rows);
  if (text->types)
    free(text->types);
  if (text->current_row_buffer)
    text_row_free(text->current_row_buffer);
  gaiaFreeUTF8Converter(text->toUtf8);
  fclose(text->text_file);
  free(text);
}

static int text_is_integer(char *value)
{
/* checking if this value can be an INTEGER */
  int invalids = 0;
  int digits = 0;
  int signs = 0;
  char last = '\0';
  char *p = value;
  while (*p != '\0')
    {
      last = *p;
      if (*p >= '0' && *p <= '9')
        digits++;
      else if (*p == '+' || *p == '-')
        signs++;
      else
        signs++;
      p++;
    }
  if (invalids)
    return 0;
  if (signs > 1)
    return 0;
  if (signs)
    {
      if (*value == '+' || *value == '-' || last == '+' || last == '-')
        ;
      else
        return 0;
    }
  return 1;
}

static int text_is_double(char *value, char decimal_separator)
{
/* checking if this value can be a DOUBLE */
  int invalids = 0;
  int digits = 0;
  int signs = 0;
  int points = 0;
  char last = '\0';
  char *p = value;
  while (*p != '\0')
    {
      last = *p;
      if (*p >= '0' && *p <= '9')
        digits++;
      else if (*p == '+' || *p == '-')
        points++;
      else
        {
          if (decimal_separator == ',')
            {
              if (*p == ',')
                points++;
              else
                invalids++;
          } else
            {
              if (*p == '.')
                points++;
              else
                invalids++;
            }
        }
      p++;
    }
  if (invalids)
    return 0;
  if (points > 1)
    return 0;
  if (signs > 1)
    return 0;
  if (signs)
    {
      if (*value == '+' || *value == '-' || last == '+' || last == '-')
        ;
      else
        return 0;
    }
  return 1;
}

static void text_clean_integer(char *value)
{
/* cleaning an integer value */
  char last;
  char buffer[35536];
  int len = strlen(value);
  last = value[len - 1];
  if (last == '-' || last == '+')
    {
      /* trailing sign; transforming into a leading sign */
      *buffer = last;
      strcpy(buffer + 1, value);
      buffer[len - 1] = '\0';
      strcpy(value, buffer);
    }
}

static void text_clean_double(char *value)
{
/* cleaning an integer value */
  char *p;
  char last;
  char buffer[35536];
  int len = strlen(value);
  last = value[len - 1];
  if (last == '-' || last == '+')
    {
      /* trailing sign; transforming into a leading sign */
      *buffer = last;
      strcpy(buffer + 1, value);
      buffer[len - 1] = '\0';
      strcpy(value, buffer);
    }
  p = value;
  while (*p != '\0')
    {
      /* transforming COMMAs into POINTs */
      if (*p == ',')
        *p = '.';
      p++;
    }
}

static int text_clean_text(char **value, void *toUtf8)
{
/* cleaning a TEXT value and converting to UTF-8 */
  char *text = *value;
  char *utf8text;
  int err;
  int i;
  int oldlen = strlen(text);
  int newlen;
  for (i = oldlen - 1; i > 0; i++)
    {
      /* cleaning up trailing spaces */
      if (text[i] == ' ')
        text[i] = '\0';
      else
        break;
    }
  utf8text = gaiaConvertToUTF8(toUtf8, text, oldlen, &err);
  if (err)
    return 1;
  newlen = strlen(utf8text);
  if (newlen <= oldlen)
    strcpy(*value, utf8text);
  else
    {
      free(*value);
      *value = (char *) malloc(newlen + 1);
      if (*value == NULL)
        {
          fprintf(stderr, "LoadText: insufficient memory\n");
          fflush(stderr);
          return 1;
        }
      strcpy(*value, utf8text);
    }
  return 0;
}

static struct row_buffer *text_create_row(char **fields, char *text_mark,
                                          int max_cell)
{
/* creating a row struct */
  int i;
  struct row_buffer *row =
    (struct row_buffer *) malloc(sizeof(struct row_buffer));
  row->n_cells = max_cell;
  if (max_cell < 0)
    row->cells = NULL;
  else
    {
      row->cells = (char **) malloc(sizeof(char *) * (max_cell));
      if (row->cells == NULL)
        {
          fprintf(stderr, "LoadText: insufficient memory\n");
          fflush(stderr);
          return NULL;
        }
      row->text = (char *) malloc(sizeof(char) * (max_cell));
      if (row->text == NULL)
        {
          fprintf(stderr, "LoadText: insufficient memory\n");
          fflush(stderr);
          return NULL;
        }
      for (i = 0; i < row->n_cells; i++)
        {
          /* setting cell values */
          *(row->cells + i) = *(fields + i);
          *(row->text + i) = *(text_mark + i);
        }
    }
  return row;
}

static struct row_buffer *text_read_row(struct text_buffer *text,
                                        struct row_pointer *ptr,
                                        int preliminary)
{
/* parsing a single row from the underlaying text file */
  int fld;
  int len;
  int max_cell;
  int is_text = 0;
  int is_string = 0;
  char last = '\0';
  char *fields[4096];
  char text_mark[4096];
  char buffer[35536];
  char *p = buffer;
  char buf_in[65536];
  char *in = buf_in;
  struct row_buffer *row;
  if (fseek(text->text_file, ptr->offset, SEEK_SET) < 0)
    {
      fprintf(stderr, "LoadText: corrupted text file\n");
      fflush(stderr);
      return NULL;
    }
  if ((int) fread(buf_in, 1, ptr->len, text->text_file) != ptr->len)
    {
      fprintf(stderr, "LoadText: corrupted text file\n");
      fflush(stderr);
      return NULL;
    }
  if (*buf_in != ptr->start)
    {
      fprintf(stderr, "LoadText: corrupted text file\n");
      fflush(stderr);
      return NULL;
    }
  if (*(buf_in + ptr->len - 1) != '\n')
    {
      fprintf(stderr, "LoadText: corrupted text file\n");
      fflush(stderr);
      return NULL;
    }
  buf_in[ptr->len] = '\0';
  for (fld = 0; fld < 4096; fld++)
    {
      /* preparing an empty row */
      fields[fld] = NULL;
      text_mark[fld] = 0;
    }
  fld = 0;
  while (*in != '\0')
    {
      /* parsing the row, one char at each time */
      if (*in == '\r' && !is_string)
        {
          last = *in++;
          continue;
        }
      if (*in == text->field_separator && !is_string)
        {
          /* inserting a field into the fields tmp array */
          last = *in;
          *p = '\0';
          len = strlen(buffer);
          if (len)
            {
              fields[fld] = (char *) malloc(len + 1);
              strcpy(fields[fld], buffer);
              text_mark[fld] = is_text;
            }
          fld++;
          p = buffer;
          *p = '\0';
          is_text = 0;
          in++;
          continue;
        }
      if (*in == text->text_separator)
        {
          /* found a text separator */
          if (is_string)
            {
              is_string = 0;
              last = *in;
              is_text = 1;
          } else
            {
              if (last == text->text_separator)
                *p++ = text->text_separator;
              is_string = 1;
            }
          in++;
          continue;
        }
      last = *in;
      if (*in == '\n' && !is_string)
        {
          /* inserting the row into the text buffer */
          *p = '\0';
          len = strlen(buffer);
          if (len)
            {
              fields[fld] = (char *) malloc(len + 1);
              strcpy(fields[fld], buffer);
            }
          fld++;
          p = buffer;
          *p = '\0';
          max_cell = -1;
          for (fld = 0; fld < 4096; fld++)
            {
              if (fields[fld])
                {
                  max_cell = fld;
                }
            }
          max_cell++;
          if (!preliminary)
            {
              if (max_cell != ptr->n_cells)
                {
                  for (fld = 0; fld < 4096; fld++)
                    {
                      if (fields[fld])
                        free(fields[fld]);
                    }
                  return NULL;
                }
            }
          row = text_create_row(fields, text_mark, max_cell);
          return row;
        }
      *p++ = *in++;
    }
  return NULL;
}

static struct text_buffer *text_parse(const char *path, const char *encoding,
                                      bool first_line_titles,
                                      const char field_separator,
                                      const char text_separator,
                                      const char decimal_separator)
{
/* trying to open and parse the text file */
  int c;
  int len;
  int fld;
  int is_string = 0;
  char col_types[4096];
  char buffer[35536];
  char *p = buffer;
  struct text_buffer *text;
  int nrows;
  int max_cols;
  int errs;
  struct row_block *block;
  struct row_buffer *row;
  void *toUtf8;
  int ir;
  char title[64];
  char *first_valid_row;
  int i;
  char *name;
  FILE *in;
  long cur_pos = 0;
  long begin_offset = 0;
  char start;
  int new_row = 1;
/* trying to open the text file */
  in = fopen(path, "rb");
  if (!in)
    {
      fprintf(stderr, "LoadText: open error '%s'\n", path);
      fflush(stderr);
      return NULL;
    }
  toUtf8 = gaiaCreateUTF8Converter(encoding);
  if (!toUtf8)
    {
      fprintf(stderr, "LoadText: illegal charset\n");
      fflush(stderr);
      return NULL;
    }
  text =
    text_buffer_alloc(in, toUtf8, field_separator, text_separator,
                      decimal_separator);
  if (text == NULL)
    {
      fprintf(stderr, "LoadText: insufficient memory\n");
      fflush(stderr);
      return NULL;
    }
  while ((c = getc(in)) != EOF)
    {
      /* parsing the file, one char at each time */
      cur_pos++;
      if (c == '\r' && !is_string)
        continue;
      if (new_row)
        {
          /* a new row begins here */
          new_row = 0;
          start = (char) c;
          begin_offset = cur_pos - 1;
          p = buffer;
        }
      if (c == text_separator)
        {
          /* found a text separator */
          if (is_string)
            is_string = 0;
          else
            is_string = 1;
          continue;
        }
      if (c == '\n' && !is_string)
        {
          /* inserting the row into the text buffer */
          len = cur_pos - begin_offset;
          if (!text_insert_row(text, begin_offset, len, start))
            {
              text_buffer_free(text);
              fprintf(stderr, "LoadText: insufficient memory\n");
              fflush(stderr);
              return NULL;
            }
          new_row = 1;
          continue;
        }
      *p++ = c;
    }
/* checking if the text file really seems to contain a table */
  nrows = 0;
  errs = 0;
  block = text->first;
  while (block)
    {
      for (i = 0; i < block->n_rows; i++)
        {
          if (first_line_titles && block == text->first && i == 0)
            {
              /* skipping first line */
              continue;
            }
          nrows++;
        }
      block = block->next;
    }
  if (nrows == 0)
    {
      text_buffer_free(text);
      return NULL;
    }
/* going to check the column types */
  first_valid_row = (char *) malloc(sizeof(char) * 4096);
  if (first_valid_row == NULL)
    {
      text_buffer_free(text);
      fprintf(stderr, "LoadText: insufficient memory\n");
      fflush(stderr);
      return NULL;
    }
  for (fld = 0; fld < 4096; fld++)
    {
      /* initally assuming any cell contains TEXT */
      *(col_types + fld) = VRTTXT_TEXT;
      *(first_valid_row + fld) = 1;
    }
  nrows = 0;
  max_cols = 0;
  block = text->first;
  while (block)
    {
      for (i = 0; i < block->n_rows; i++)
        {
          row = text_read_row(text, block->rows + i, 1);
          if (row == NULL)
            {
              text_buffer_free(text);
              return NULL;
            }
          if (first_line_titles && block == text->first && i == 0)
            {
              /* skipping first line */
              for (fld = 0; fld < row->n_cells; fld++)
                {
                  if (*(row->cells + fld))
                    {
                      if (fld > max_cols)
                        max_cols = fld;
                    }
                }
              (block->rows + i)->valid = 1;
              (block->rows + i)->n_cells = row->n_cells;
              text_row_free(row);
              continue;
            }
          for (fld = 0; fld < row->n_cells; fld++)
            {
              if (*(row->cells + fld))
                {
                  if ((fld + 1) > max_cols)
                    max_cols = fld + 1;
                  if (text_is_integer(*(row->cells + fld))
                      && !(*(row->text + fld)))
                    {
                      if (*(first_valid_row + fld))
                        {
                          *(col_types + fld) = VRTTXT_INTEGER;
                          *(first_valid_row + fld) = 0;
                        }
                  } else if (text_is_double
                             (*(row->cells + fld), decimal_separator)
                             && !(*(row->text + fld)))
                    {
                      if (*(first_valid_row + fld))
                        {
                          *(col_types + fld) = VRTTXT_DOUBLE;
                          *(first_valid_row + fld) = 0;
                      } else
                        {
                          /* promoting an INTEGER column to be of the DOUBLE type */
                          if (*(col_types + fld) == VRTTXT_INTEGER)
                            *(col_types + fld) = VRTTXT_DOUBLE;
                        }
                  } else
                    {
                      /* this column is anyway of the TEXT type */
                      *(col_types + fld) = VRTTXT_TEXT;
                      if (*(first_valid_row + fld))
                        *(first_valid_row + fld) = 0;
                    }
                }
            }
          nrows++;
          (block->rows + i)->valid = 1;
          (block->rows + i)->n_cells = row->n_cells;
          text_row_free(row);
        }
      block = block->next;
    }
  free(first_valid_row);
  if (nrows == 0 && max_cols == 0)
    {
      text_buffer_free(text);
      return NULL;
    }
  text->n_rows = nrows;
  text->max_n_cells = max_cols;
  text->types = (char *) malloc(sizeof(char) * text->max_n_cells);
  if (text->types == NULL)
    {
      text_buffer_free(text);
      fprintf(stderr, "LoadText: insufficient memory\n");
      fflush(stderr);
      return NULL;
    }
  memcpy(text->types, col_types, text->max_n_cells);
/* preparing the column names */
  text->titles = (char **) malloc(sizeof(char *) * text->max_n_cells);
  if (text->titles == NULL)
    {
      text_buffer_free(text);
      fprintf(stderr, "LoadText: insufficient memory\n");
      fflush(stderr);
      return NULL;
    }
  block = text->first;
  if (first_line_titles && block)
    {
      row = text_read_row(text, block->rows + 0, 0);
      if (row == NULL)
        {
          text_buffer_free(text);
          return NULL;
        }
      for (fld = 0; fld < text->max_n_cells; fld++)
        {
          if (fld >= row->n_cells)
            {
              /* this column name is NULL; setting a default name */
              sprintf(title, "COL%03d", fld + 1);
              len = strlen(title);
              *(text->titles + fld) = (char *) malloc(len + 1);
              if (*(text->titles + fld) == NULL)
                {
                  text_buffer_free(text);
                  fprintf(stderr, "LoadText: insufficient memory\n");
                  fflush(stderr);
                  return NULL;
                }
              strcpy(*(text->titles + fld), title);
          } else
            {
              if (*(row->cells + fld))
                {
                  len = strlen(*(row->cells + fld));
                  *(text->titles + fld) = (char *) malloc(len + 1);
                  if (*(text->titles + fld) == NULL)
                    {
                      text_buffer_free(text);
                      fprintf(stderr, "LoadText: insufficient memory\n");
                      fflush(stderr);
                      return NULL;
                    }
                  strcpy(*(text->titles + fld), *(row->cells + fld));
                  name = *(text->titles + fld);
                  for (i = 0; i < len; i++)
                    {
                      /* masking any space in the column name */
                      if (*(name + i) == ' ')
                        *(name + i) = '_';
                    }
              } else
                {
                  /* this column name is NULL; setting a default name */
                  sprintf(title, "COL%03d", fld + 1);
                  len = strlen(title);
                  *(text->titles + fld) = (char *) malloc(len + 1);
                  strcpy(*(text->titles + fld), title);
                }
            }
        }
      text_row_free(row);
  } else
    {
      for (fld = 0; fld < text->max_n_cells; fld++)
        {
          sprintf(title, "COL%03d", fld + 1);
          len = strlen(title);
          *(text->titles + fld) = (char *) malloc(len + 1);
          if (*(text->titles + fld) == NULL)
            {
              text_buffer_free(text);
              fprintf(stderr, "LoadText: insufficient memory\n");
              fflush(stderr);
              return NULL;
            }
          strcpy(*(text->titles + fld), title);
        }
    }
/* ok, we can now go to prepare the rows array */
  text->rows =
    (struct row_pointer **) malloc(sizeof(struct row_pointer *) * text->n_rows);
  if (text->rows == NULL)
    {
      text_buffer_free(text);
      fprintf(stderr, "LoadText: insufficient memory\n");
      fflush(stderr);
      return NULL;
    }
  ir = 0;
  block = text->first;
  while (block)
    {
      for (i = 0; i < block->n_rows; i++)
        {
          if (first_line_titles && block == text->first && i == 0)
            {
              /* skipping first line */
              continue;
            }
          if ((block->rows + i)->valid)
            *(text->rows + ir++) = block->rows + i;
        }
      block = block->next;
    }
  return text;
}

void
  MyFrame::LoadText(wxString & path, wxString & table, wxString & charset,
                    bool first_titles, const char decimal_separator,
                    const char separator, const char text_separator)
{
//
// loading a CSV/TXT as a new DB table
//
  struct text_buffer *text = NULL;
  struct row_buffer *row;
  struct row_pointer *pRow;
  int seed;
  int dup;
  int idup;
  char dummy[65536];
  char dummyName[4096];
  char **col_name = NULL;
  int i;
  char sql[65536];
  int len;
  int ret;
  int rows = 0;
  char *errMsg = NULL;
  bool sqlError = false;
  char xtable[1024];
  int current_row;
  wxString msg;
  char buf[4096];
//
// performing some checks before starting
//
  if (TableAlreadyExists(table) == true)
    {
      wxMessageBox(wxT("a table name '") + table + wxT("' already exists"),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      return;
    }
  text =
    text_parse(path.ToUTF8(), charset.ToUTF8(), first_titles, separator,
               text_separator, decimal_separator);
  if (!text)
    return;
  ::wxBeginBusyCursor();
//
// checking for duplicate / illegal column names and antialising them 
//
  col_name = (char **) malloc(sizeof(char *) * text->max_n_cells);
  seed = 0;
  for (i = 0; i < text->max_n_cells; i++)
    {
      strcpy(dummyName, *(text->titles + i));
      dup = 0;
      for (idup = 0; idup < i; idup++)
        {
          if (strcasecmp(dummyName, *(col_name + idup)) == 0)
            dup = 1;
        }
      if (strcasecmp(dummyName, "PKUID") == 0)
        dup = 1;
      if (strcasecmp(dummyName, "Geometry") == 0)
        dup = 1;
      if (dup)
        sprintf(dummyName, "COL_%d", seed++);
      len = strlen(dummyName);
      *(col_name + i) = (char *) malloc(len + 1);
      strcpy(*(col_name + i), dummyName);
    }
//
// starting a transaction
//
  ret = sqlite3_exec(SqliteHandle, "BEGIN", NULL, 0, &errMsg);
  if (ret != SQLITE_OK)
    {
      wxMessageBox(wxT("load CSV/TXT error:") + wxString::FromUTF8(errMsg),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      sqlite3_free(errMsg);
      sqlError = true;
      goto clean_up;
    }
//
// creating the Table 
//
  strcpy(xtable, table.ToUTF8());
  sprintf(sql, "CREATE TABLE \"%s\"", xtable);
  strcat(sql, " (\n\"PK_UID\" INTEGER PRIMARY KEY AUTOINCREMENT");
  for (i = 0; i < text->max_n_cells; i++)
    {
      strcat(sql, ",\n\"");
      strcat(sql, *(col_name + i));
      if (*(text->types + i) == VRTTXT_INTEGER)
        strcat(sql, "\" INTEGER");
      else if (*(text->types + i) == VRTTXT_DOUBLE)
        strcat(sql, "\" DOUBLE");
      else
        strcat(sql, "\" TEXT");
    }
  strcat(sql, ")");
  ret = sqlite3_exec(SqliteHandle, sql, NULL, 0, &errMsg);
  if (ret != SQLITE_OK)
    {
      wxMessageBox(wxT("load text error:") + wxString::FromUTF8(errMsg),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      sqlite3_free(errMsg);
      sqlError = true;
      goto clean_up;
    }
  current_row = 0;
  while (current_row < text->n_rows)
    {
      //
      // inserting rows from CSV/TXT
      //
      pRow = get_row_pointer(text, current_row);
      if (pRow)
        {
          row = text_read_row(text, pRow, 0);
          if (row == NULL)
            break;
        }
      sprintf(sql, "INSERT INTO \"%s\" (\n\"PK_UID\"", xtable);
      for (i = 0; i < text->max_n_cells; i++)
        {
          // columns corresponding to some CSV/TXT column
          strcat(sql, ",\"");
          strcat(sql, *(col_name + i));
          strcat(sql, "\"");
        }
      strcat(sql, ")\nVALUES (");
      sprintf(dummy, "%d", current_row);
      strcat(sql, dummy);
      for (i = 0; i < text->max_n_cells; i++)
        {
          // column values
          strcat(sql, ",");
          if (i >= row->n_cells)
            strcat(sql, "NULL");
          else
            {
              if (*(row->cells + i))
                {
                  if (*(text->types + i) == VRTTXT_INTEGER)
                    {
                      strcpy(buf, *(row->cells + i));
                      text_clean_integer(buf);
                      sprintf(dummy, "%d", atoi(buf));
                      strcat(sql, dummy);
                  } else if (*(text->types + i) == VRTTXT_DOUBLE)
                    {
                      strcpy(buf, *(row->cells + i));
                      text_clean_double(buf);
                      sprintf(dummy, "%1.6f", atof(buf));
                      strcat(sql, dummy);
                  } else
                    {
                      if (text_clean_text(row->cells + i, text->toUtf8))
                        {
                          fprintf(stderr, "LoadText: encoding error\n");
                          fflush(stderr);
                        }
                      strcpy(dummy, *(row->cells + i));
                      CleanSqlString(dummy);
                      strcat(sql, "'");
                      strcat(sql, dummy);
                      strcat(sql, "'");
                    }
              } else
                strcat(sql, "NULL");
            }
        }
      strcat(sql, ")");
      ret = sqlite3_exec(SqliteHandle, sql, NULL, 0, &errMsg);
      if (ret != SQLITE_OK)
        {
          wxMessageBox(wxT("load text error:") + wxString::FromUTF8(errMsg),
                       wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
          sqlite3_free(errMsg);
          sqlError = true;
          goto clean_up;
        }
      text_row_free(row);
      rows++;
      current_row++;
    }
clean_up:
  if (col_name)
    {
      // releasing memory allocation for column names 
      for (i = 0; i < text->max_n_cells; i++)
        free(*(col_name + i));
      free(col_name);
    }
  free(text);
  if (sqlError == true)
    {
      // some error occurred - ROLLBACK 
      ret = sqlite3_exec(SqliteHandle, "ROLLBACK", NULL, 0, &errMsg);
      if (ret != SQLITE_OK)
        {
          wxMessageBox(wxT("load text error:") + wxString::FromUTF8(errMsg),
                       wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
          sqlite3_free(errMsg);
        }
      ::wxEndBusyCursor();
      msg =
        wxT("CSV/TXT not loaded\n\n\na ROLLBACK was automatically performed");
      wxMessageBox(msg, wxT("spatialite-gui"), wxOK | wxICON_WARNING, this);
  } else
    {
      // ok - confirming pending transaction - COMMIT 
      ret = sqlite3_exec(SqliteHandle, "COMMIT", NULL, 0, &errMsg);
      if (ret != SQLITE_OK)
        {
          wxMessageBox(wxT("load text error:") + wxString::FromUTF8(errMsg),
                       wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
          sqlite3_free(errMsg);
          return;
        }
      ::wxEndBusyCursor();
      sprintf(dummy, "CSV/TXT loaded\n\n%d inserted rows", rows);
      msg = wxString::FromUTF8(dummy);
      wxMessageBox(msg, wxT("spatialite-gui"), wxOK | wxICON_INFORMATION, this);
      InitTableTree();
    }
}

void MyFrame::LoadDbf(wxString & path, wxString & table, wxString & charset,
                      bool silent)
{
//
// loading a DBF as a new DB table
//
  sqlite3_stmt *stmt;
  int cnt;
  int col_cnt;
  int seed;
  int len;
  int dup;
  int idup;
  int current_row;
  int ret;
  int deleted;
  int rows = 0;
  char *errMsg = NULL;
  char xtable[1024];
  char dummyName[4096];
  char sql[65536];
  char **col_name = NULL;
  wxString dummyStr;
  wxString msg;
  gaiaDbfPtr dbf = NULL;
  gaiaDbfFieldPtr dbf_field;
  bool sqlError = false;
//
// performing some checks before starting
//
  if (TableAlreadyExists(table) == true)
    {
      wxMessageBox(wxT("a table name '") + table + wxT("' already exists"),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      return;
    }
//
// initalizing the DBF struct
//
  dbf = gaiaAllocDbf();
  gaiaOpenDbfRead(dbf, path.ToUTF8(), charset.ToUTF8(), "UTF-8");
  if (!(dbf->Valid))
    {
      wxString error = wxT("ERROR: invalid DBF\n\n");
      if (dbf->LastError)
        error += wxString::FromUTF8(dbf->LastError);
      gaiaFreeDbf(dbf);
      wxMessageBox(error, wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      return;
    }
  ::wxBeginBusyCursor();
//
// checking for duplicate / illegal column names and antialising them 
//
  col_cnt = 0;
  dbf_field = dbf->Dbf->First;
  while (dbf_field)
    {
      // counting DBF fields
      col_cnt++;
      dbf_field = dbf_field->Next;
    }
  col_name = (char **) malloc(sizeof(char *) * col_cnt);
  cnt = 0;
  seed = 0;
  dbf_field = dbf->Dbf->First;
  while (dbf_field)
    {
      // preparing column names
      strcpy(dummyName, dbf_field->Name);
      dup = 0;
      for (idup = 0; idup < cnt; idup++)
        {
          if (strcasecmp(dummyName, *(col_name + idup)) == 0)
            dup = 1;
        }
      if (strcasecmp(dummyName, "PK_UID") == 0)
        dup = 1;
      if (dup)
        sprintf(dummyName, "COL_%d", seed++);
      len = strlen(dummyName);
      *(col_name + cnt) = (char *) malloc(len + 1);
      strcpy(*(col_name + cnt), dummyName);
      cnt++;
      dbf_field = dbf_field->Next;
    }
//
// starting a transaction
//
  ret = sqlite3_exec(SqliteHandle, "BEGIN", NULL, 0, &errMsg);
  if (ret != SQLITE_OK)
    {
      wxMessageBox(wxT("load dbf error:") + wxString::FromUTF8(errMsg),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      sqlite3_free(errMsg);
      sqlError = true;
      goto clean_up;
    }
//
// creating the Table 
//
  strcpy(xtable, table.ToUTF8());
  sprintf(sql, "CREATE TABLE \"%s\"", xtable);
  strcat(sql, " (\n\"PK_UID\" INTEGER PRIMARY KEY AUTOINCREMENT");
  cnt = 0;
  dbf_field = dbf->Dbf->First;
  while (dbf_field)
    {
      strcat(sql, ",\n\"");
      strcat(sql, *(col_name + cnt));
      cnt++;
      switch (dbf_field->Type)
        {
          case 'C':
            strcat(sql, "\" TEXT");
            break;
          case 'N':
            if (dbf_field->Decimals)
              strcat(sql, "\" DOUBLE");
            else
              {
                if (dbf_field->Length <= 18)
                  strcat(sql, "\" INTEGER");
                else
                  strcat(sql, "\" DOUBLE");
              }
            break;
          case 'D':
            strcat(sql, "\" DOUBLE");
            break;
          case 'F':
            strcat(sql, "\" DOUBLE");
            break;
          case 'L':
            strcat(sql, "\" INTEGER");
            break;
        };
      dbf_field = dbf_field->Next;
    }
  strcat(sql, ")");
  ret = sqlite3_exec(SqliteHandle, sql, NULL, 0, &errMsg);
  if (ret != SQLITE_OK)
    {
      wxMessageBox(wxT("load dbf error:") + wxString::FromUTF8(errMsg),
                   wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
      sqlite3_free(errMsg);
      sqlError = true;
      goto clean_up;
    }
// preparing the INSERT INTO parameterized statement
  sprintf(sql, "INSERT INTO \"%s\" (\"PK_UID\"", xtable);
  cnt = 0;
  dbf_field = dbf->Dbf->First;
  while (dbf_field)
    {
      // columns corresponding to some DBF attribute 
      strcat(sql, ",\"");
      strcat(sql, *(col_name + cnt));
      strcat(sql, "\"");
      cnt++;
      dbf_field = dbf_field->Next;
    }
  strcat(sql, ")\nVALUES (? ");
  dbf_field = dbf->Dbf->First;
  while (dbf_field)
    {
      // column values
      strcat(sql, ", ?");
      dbf_field = dbf_field->Next;
    }
  strcat(sql, ")");
  ret = sqlite3_prepare_v2(SqliteHandle, sql, strlen(sql), &stmt, NULL);
  if (ret != SQLITE_OK)
    {
      wxString err = wxString::FromUTF8(sqlite3_errmsg(SqliteHandle));
      wxMessageBox(wxT("load dbf error: ") + err, wxT("spatialite-gui"),
                   wxOK | wxICON_ERROR, this);
      sqlError = true;
      goto clean_up;
    }
  current_row = 0;
  while (1)
    {
      //
      // inserting rows from DBF
      //
      ret = gaiaReadDbfEntity(dbf, current_row, &deleted);
      if (!ret)
        {
          if (!(dbf->LastError))  // normal DBF EOF
            break;
          wxMessageBox(wxT("load dbf error:") +
                       wxString::FromUTF8(dbf->LastError),
                       wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
          sqlError = true;
          goto clean_up;
        }
      current_row++;
      if (deleted)
        {
          // skipping DBF deleted row
          continue;
        }
      // binding query params
      sqlite3_reset(stmt);
      sqlite3_clear_bindings(stmt);
      sqlite3_bind_int(stmt, 1, current_row);
      cnt = 0;
      dbf_field = dbf->Dbf->First;
      while (dbf_field)
        {
          // column values
          if (!(dbf_field->Value))
            sqlite3_bind_null(stmt, cnt + 2);
          else
            {
              switch (dbf_field->Value->Type)
                {
                  case GAIA_INT_VALUE:
                    sqlite3_bind_int64(stmt, cnt + 2,
                                       dbf_field->Value->IntValue);
                    break;
                  case GAIA_DOUBLE_VALUE:
                    sqlite3_bind_double(stmt, cnt + 2,
                                        dbf_field->Value->DblValue);
                    break;
                  case GAIA_TEXT_VALUE:
                    sqlite3_bind_text(stmt, cnt + 2,
                                      dbf_field->Value->TxtValue,
                                      strlen(dbf_field->Value->
                                             TxtValue), SQLITE_STATIC);
                    break;
                  default:
                    sqlite3_bind_null(stmt, cnt + 2);;
                    break;
                };
            }
          cnt++;
          dbf_field = dbf_field->Next;
        }
      ret = sqlite3_step(stmt);
      if (ret == SQLITE_DONE || ret == SQLITE_ROW)
        ;
      else
        {
          wxString err = wxString::FromUTF8(sqlite3_errmsg(SqliteHandle));
          wxMessageBox(wxT("load dbf error:") + err,
                       wxT("spatialite-gui"), wxOK | wxICON_ERROR, this);
          sqlite3_finalize(stmt);
          sqlError = true;
          goto clean_up;
        }
      rows++;
    }
  sqlite3_finalize(stmt);
clean_up:
  gaiaFreeDbf(dbf);
  if (col_name)
    {
      // releasing memory allocation for column names 
      for (cnt = 0; cnt < col_cnt; cnt++)
        free(*(col_name + cnt));
      free(col_name);
    }
  if (sqlError == true)
    {
      // some error occurred - ROLLBACK 
      ret = sqlite3_exec(SqliteHandle, "ROLLBACK", NULL, 0, &errMsg);
      if (ret != SQLITE_OK)
        {
          wxMessageBox(wxT("load dbf error:") +
                       wxString::FromUTF8(errMsg), wxT("spatialite-gui"),
                       wxOK | wxICON_ERROR, this);
          sqlite3_free(errMsg);
        }
      ::wxEndBusyCursor();
      msg = wxT("DBF not loaded\n\n\na ROLLBACK was automatically performed");
      wxMessageBox(msg, wxT("spatialite-gui"), wxOK | wxICON_WARNING, this);
  } else
    {
      // ok - confirming pending transaction - COMMIT 
      ret = sqlite3_exec(SqliteHandle, "COMMIT", NULL, 0, &errMsg);
      if (ret != SQLITE_OK)
        {
          wxMessageBox(wxT("load dbf error:") +
                       wxString::FromUTF8(errMsg), wxT("spatialite-gui"),
                       wxOK | wxICON_ERROR, this);
          sqlite3_free(errMsg);
          return;
        }
      ::wxEndBusyCursor();
      if (silent == false)
        {
          sprintf(dummyName, "DBF loaded\n\n%d inserted rows", rows);
          msg = wxString::FromUTF8(dummyName);
          wxMessageBox(msg, wxT("spatialite-gui"), wxOK | wxICON_INFORMATION,
                       this);
        }
      InitTableTree();
    }
}
